use chrono::NaiveTime;
use garde::Validate;

use crate::{END_OF_DAY, START_OF_DAY};

use super::{start_time_lies_before_end_time, Availability, AvailabilityTime};

#[derive(Validate)]
#[garde(allow_unvalidated)]
#[garde(context(AvailabilityContext))]
pub struct AvailabilityChangeset {
    #[garde(
        custom(time_is_not_already_made_available),
        custom(temporarily_start_time_before_end_time)
    )]
    pub time: AvailabilityTime,
    pub comment: Option<String>,
}

pub struct AvailabilityContext {
    pub existing_availabilities: Vec<Availability>,
}

pub fn only_one_availability_exists_and_is_whole_day(a: &[Availability]) -> bool {
    a.len() == 1 && a[0].time == AvailabilityTime::WholeDay
}

fn time_is_not_already_made_available(
    value: &AvailabilityTime,
    context: &AvailabilityContext,
) -> garde::Result {
    if let AvailabilityTime::Temporarily(_, _) = value {
        if only_one_availability_exists_and_is_whole_day(&context.existing_availabilities) {
            return Err(garde::Error::new("cant create a availability while an other availability for the whole day is already present"));
        }

        if find_free_time_slots(&context.existing_availabilities).len() == 0 {
            return Err(garde::Error::new(
                "cant create a availability as every time slot is already filled",
            ));
        }
    } else {
        if context.existing_availabilities.len() > 0 {
            return Err(garde::Error::new("cant create a availability for the whole day while an other availability is already present"));
        }
    }

    Ok(())
}

fn temporarily_start_time_before_end_time(
    value: &AvailabilityTime,
    context: &AvailabilityContext,
) -> garde::Result {
    if let AvailabilityTime::Temporarily(start, end) = value {
        return start_time_lies_before_end_time(&(*start, *end), context);
    }

    Ok(())
}

pub fn find_free_time_slots(availabilities: &[Availability]) -> Vec<(NaiveTime, NaiveTime)> {
    let mut times = Vec::new();

    for a in availabilities {
        if let AvailabilityTime::Temporarily(start, end) = a.time {
            times.push((start, end));
        }
    }

    if times.len() == 0 {
        return Vec::new();
    }
    //println!("zeiten {times:?}");

    times.sort();

    //println!("zeiten sort {times:?}");

    let mut changed = true;
    while changed {
        changed = false;

        for i in 0..(times.len() - 1) {
            let b = times[i + 1].clone();
            let a = times.get_mut(i).unwrap();

            if a.1 == b.0 {
                a.1 = b.1;
                times.remove(i + 1);
                changed = true;
                break;
            }
        }
    }

    //println!("zeiten unified {times:?}");

    // now times contains unified list of existing availabilities -> now calculate the "inverse"

    let mut available_slots = Vec::new();
    let start = times.first().unwrap();
    if start.0 != START_OF_DAY {
        available_slots.push((START_OF_DAY, start.0));
    }

    let mut iterator = times.iter().peekable();
    while let Some(a) = iterator.next() {
        if let Some(b) = iterator.peek() {
            available_slots.push((a.1, b.0));
        }
    }

    let end = times.last().unwrap();
    if end.1 != END_OF_DAY {
        available_slots.push((end.1, END_OF_DAY));
    }

    //println!("available {available_slots:?}");
    available_slots
}
