use chrono::NaiveDate;
use chrono::NaiveTime;
use garde::Validate;

use super::start_time_lies_before_end_time;

#[derive(Validate)]
#[garde(allow_unvalidated)]
#[garde(context(Option<EventContext> as ctx))]
pub struct EventChangeset {
    #[garde(custom(date_unchanged_if_edit))]
    pub date: NaiveDate,
    #[garde(
        custom(start_time_lies_before_end_time),
        custom(time_can_be_extended_if_edit)
    )]
    pub time: (NaiveTime, NaiveTime),
    pub name: String,
    /// check before: must exist and user can create event for this location
    pub location_id: i32,
    #[garde(custom(can_unset_wachhabender))]
    pub voluntary_wachhabender: bool,
    #[garde(custom(can_unset_fuehrungsassistent))]
    pub voluntary_fuehrungsassistent: bool,
    #[garde(range(min = ctx.as_ref().and_then(|c: &EventContext| Some(c.amount_of_assigned_posten)).unwrap_or(1), max = 100))]
    pub amount_of_posten: i16,
    pub clothing: String,
    pub note: Option<String>,
}

pub struct EventContext {
    pub date_in_db: NaiveDate,
    pub common_min_max_available_time: (NaiveTime, NaiveTime),
    pub wachhabender_assigned: bool,
    pub fuehrungsassistent_assigned: bool,
    pub amount_of_assigned_posten: i16,
}

fn date_unchanged_if_edit(value: &NaiveDate, context: &Option<EventContext>) -> garde::Result {
    if context.as_ref().is_some_and(|c| c.date_in_db != *value) {
        return Err(garde::Error::new("event date can't be changed"));
    }

    Ok(())
}

fn time_can_be_extended_if_edit(
    value: &(NaiveTime, NaiveTime),
    context: &Option<EventContext>,
) -> garde::Result {
    if let Some(context) = context {
        let old_start_time = context.common_min_max_available_time.0;
        let new_start_time = value.0;
        let old_end_time = context.common_min_max_available_time.1;
        let new_end_time = value.1;

        if new_start_time < old_start_time {
            return Err(garde::Error::new(
                "starttime lies outside of available time for assigned people",
            ));
        }

        if new_end_time > old_end_time {
            return Err(garde::Error::new(
                "endtime lies ouside of available time for assigned people",
            ));
        }
    }

    Ok(())
}

fn can_unset_fuehrungsassistent(value: &bool, context: &Option<EventContext>) -> garde::Result {
    if context
        .as_ref()
        .is_some_and(|c| !*value && c.fuehrungsassistent_assigned)
    {
        return Err(garde::Error::new(
            "fuehrungsassistent can't be set to not by ff, because a person is already assigned",
        ));
    }

    Ok(())
}

fn can_unset_wachhabender(value: &bool, context: &Option<EventContext>) -> garde::Result {
    if context
        .as_ref()
        .is_some_and(|c| !*value && c.wachhabender_assigned)
    {
        return Err(garde::Error::new(
            "wachhabender can't be set to not by ff, because a person is already assigned",
        ));
    }

    Ok(())
}
